@@ -1,12 +1,18 @@
-/* session.c -- Session management for libcoap
-*
-* Copyright (C) 2017 Jean-Claue Michelou <jcm@spinetix.com>
-*
+/* coap_session.c -- Session management for libcoap
+ *
+ * Copyright (C) 2017 Jean-Claue Michelou <jcm@spinetix.com>
+ * Copyright (C) 2022 Jon Shallow <supjps-libcoap@jpshallow.com>
+ *
  * SPDX-License-Identifier: BSD-2-Clause
  *
-* This file is part of the CoAP library libcoap. Please see
-* README for terms of use.
-*/
+ * This file is part of the CoAP library libcoap. Please see
+ * README for terms of use.
+ */
+
+/**
+ * @file coap_session.c
+ * @brief Session handling functions
+ */
 
 #include "coap3/coap_internal.h"
 
@@ -21,51 +27,101 @@
 #endif /* COAP_EPOLL_SUPPORT */
 #include <errno.h>
 
-void
-coap_session_set_max_retransmit (coap_session_t *session, unsigned int value) {
-  if (value > 0)
-    session->max_retransmit = value;
-  coap_log(LOG_DEBUG, "***%s: session max_retransmit set to %d\n",
-           coap_session_str(session), session->max_retransmit);
-  return;
-}
+#ifdef HAVE_INTTYPES_H
+#include <inttypes.h>
+#else /* ! HAVE_INTTYPES_H */
+#define PRIu32 "u"
+#endif /* ! HAVE_INTTYPES_H */
 
 void
-coap_session_set_ack_timeout (coap_session_t *session, coap_fixed_point_t value) {
-  if (value.integer_part > 0 && value.fractional_part < 1000)
+coap_session_set_ack_timeout(coap_session_t *session, coap_fixed_point_t value) {
+  if (value.integer_part > 0 && value.fractional_part < 1000) {
     session->ack_timeout = value;
-  coap_log(LOG_DEBUG, "***%s: session ack_timeout set to %d.%03d\n",
+    coap_log(LOG_DEBUG, "***%s: session ack_timeout set to %u.%03u\n",
            coap_session_str(session), session->ack_timeout.integer_part,
            session->ack_timeout.fractional_part);
-  return;
+  }
 }
 
 void
-coap_session_set_ack_random_factor (coap_session_t *session,
-                                    coap_fixed_point_t value) {
-  if (value.integer_part > 0 && value.fractional_part < 1000)
+coap_session_set_ack_random_factor(coap_session_t *session,
+                                   coap_fixed_point_t value) {
+  if (value.integer_part > 0 && value.fractional_part < 1000) {
     session->ack_random_factor = value;
-  coap_log(LOG_DEBUG, "***%s: session ack_random_factor set to %d.%03d\n",
+    coap_log(LOG_DEBUG, "***%s: session ack_random_factor set to %u.%03u\n",
            coap_session_str(session), session->ack_random_factor.integer_part,
            session->ack_random_factor.fractional_part);
-  return;
+  }
 }
 
-unsigned int
-coap_session_get_max_retransmit (const coap_session_t *session) {
-  return session->max_retransmit;
+void
+coap_session_set_max_retransmit(coap_session_t *session, uint16_t value) {
+  if (value > 0) {
+    session->max_retransmit = value;
+    coap_log(LOG_DEBUG, "***%s: session max_retransmit set to %u\n",
+           coap_session_str(session), session->max_retransmit);
+  }
+}
+
+void
+coap_session_set_nstart(coap_session_t *session, uint16_t value) {
+  if (value > 0) {
+    session->nstart = value;
+    coap_log(LOG_DEBUG, "***%s: session nstart set to %u\n",
+           coap_session_str(session), session->nstart);
+  }
+}
+
+void
+coap_session_set_default_leisure(coap_session_t *session,
+                                 coap_fixed_point_t value) {
+  if (value.integer_part > 0 && value.fractional_part < 1000) {
+    session->default_leisure = value;
+    coap_log(LOG_DEBUG, "***%s: session default_leisure set to %u.%03u\n",
+           coap_session_str(session), session->default_leisure.integer_part,
+           session->default_leisure.fractional_part);
+  }
+}
+
+void
+coap_session_set_probing_rate(coap_session_t *session, uint32_t value) {
+  if (value > 0) {
+    session->probing_rate = value;
+    coap_log(LOG_DEBUG, "***%s: session probing_rate set to %" PRIu32 "\n",
+           coap_session_str(session), session->probing_rate);
+  }
 }
 
 coap_fixed_point_t
-coap_session_get_ack_timeout (const coap_session_t *session) {
+coap_session_get_ack_timeout(const coap_session_t *session) {
   return session->ack_timeout;
 }
 
 coap_fixed_point_t
-coap_session_get_ack_random_factor (const coap_session_t *session) {
+coap_session_get_ack_random_factor(const coap_session_t *session) {
   return session->ack_random_factor;
 }
 
+uint16_t
+coap_session_get_max_retransmit(const coap_session_t *session) {
+  return session->max_retransmit;
+}
+
+uint16_t
+coap_session_get_nstart(const coap_session_t *session) {
+  return session->nstart;
+}
+
+coap_fixed_point_t
+coap_session_get_default_leisure(const coap_session_t *session) {
+  return session->default_leisure;
+}
+
+uint32_t
+coap_session_get_probing_rate(const coap_session_t *session) {
+  return session->probing_rate;
+}
+
 coap_session_t *
 coap_session_reference(coap_session_t *session) {
   ++session->ref;
@@ -108,10 +164,15 @@ coap_session_get_app_data(const coap_session_t *session) {
 
 static coap_session_t *
 coap_make_session(coap_proto_t proto, coap_session_type_t type,
-  const coap_addr_hash_t *addr_hash, const coap_address_t *local_addr,
-  const coap_address_t *remote_addr, int ifindex, coap_context_t *context,
-  coap_endpoint_t *endpoint) {
-  coap_session_t *session = (coap_session_t*)coap_malloc_type(COAP_SESSION, sizeof(coap_session_t));
+                  const coap_addr_hash_t *addr_hash,
+                  const coap_address_t *local_addr,
+                  const coap_address_t *remote_addr, int ifindex,
+                  coap_context_t *context, coap_endpoint_t *endpoint) {
+  coap_session_t *session = (coap_session_t*)coap_malloc_type(COAP_SESSION,
+                                                     sizeof(coap_session_t));
+#if ! COAP_SERVER_SUPPORT
+  (void)endpoint;
+#endif /* ! COAP_SERVER_SUPPORT */
   if (!session)
     return NULL;
   memset(session, 0, sizeof(*session));
@@ -131,12 +192,14 @@ coap_make_session(coap_proto_t proto, coap_session_type_t type,
     coap_address_init(&session->addr_info.remote);
   session->ifindex = ifindex;
   session->context = context;
+#if COAP_SERVER_SUPPORT
   session->endpoint = endpoint;
-  session->block_mode = context->block_mode;
   if (endpoint)
     session->mtu = endpoint->default_mtu;
   else
+#endif /* COAP_SERVER_SUPPORT */
     session->mtu = COAP_DEFAULT_MTU;
+  session->block_mode = context->block_mode;
   if (proto == COAP_PROTO_DTLS) {
     session->tls_overhead = 29;
     if (session->tls_overhead >= session->mtu) {
@@ -144,28 +207,34 @@ coap_make_session(coap_proto_t proto, coap_session_type_t type,
       coap_log(LOG_ERR, "DTLS overhead exceeds MTU\n");
     }
   }
-  session->max_retransmit = COAP_DEFAULT_MAX_RETRANSMIT;
   session->ack_timeout = COAP_DEFAULT_ACK_TIMEOUT;
   session->ack_random_factor = COAP_DEFAULT_ACK_RANDOM_FACTOR;
+  session->max_retransmit = COAP_DEFAULT_MAX_RETRANSMIT;
+  session->nstart = COAP_DEFAULT_NSTART;
+  session->default_leisure = COAP_DEFAULT_DEFAULT_LEISURE;
+  session->probing_rate = COAP_DEFAULT_PROBING_RATE;
   session->dtls_event = -1;
   session->last_ping_mid = COAP_INVALID_MID;
+  session->last_ack_mid = COAP_INVALID_MID;
+  session->last_con_mid = COAP_INVALID_MID;
 
-  /* initialize message id */
+  /* Randomly initialize */
   coap_prng((unsigned char *)&session->tx_mid, sizeof(session->tx_mid));
+  coap_prng((unsigned char *)&session->tx_rtag, sizeof(session->tx_rtag));
 
   return session;
 }
 
 void coap_session_mfree(coap_session_t *session) {
   coap_queue_t *q, *tmp;
-  coap_cache_entry_t *cp, *ctmp;
   coap_lg_xmit_t *lq, *ltmp;
+
+#if COAP_CLIENT_SUPPORT
   coap_lg_crcv_t *cq, *etmp;
-  coap_lg_srcv_t *sq, *stmp;
 
   /* Need to do this before (D)TLS and socket is closed down */
   LL_FOREACH_SAFE(session->lg_crcv, cq, etmp) {
-    if (cq->observe_set) {
+    if (cq->observe_set && session->no_observe_cancel == 0) {
       /* Need to close down observe */
       if (coap_cancel_observe(session, cq->app_token, COAP_MESSAGE_NON)) {
         /* Need to delete node we set up for NON */
@@ -183,6 +252,7 @@ void coap_session_mfree(coap_session_t *session) {
     LL_DELETE(session->lg_crcv, cq);
     coap_block_delete_lg_crcv(session, cq);
   }
+#endif /* COAP_CLIENT_SUPPORT */
 
   if (session->partial_pdu)
     coap_delete_pdu(session->partial_pdu);
@@ -201,12 +271,15 @@ void coap_session_mfree(coap_session_t *session) {
   if (session->psk_hint)
     coap_free(session->psk_hint);
 
+#if COAP_SERVER_SUPPORT
+  coap_cache_entry_t *cp, *ctmp;
   HASH_ITER(hh, session->context->cache, cp, ctmp) {
     /* cp->session is NULL if not session based */
     if (cp->session == session) {
       coap_delete_cache_entry(session->context, cp);
     }
   }
+#endif /* COAP_SERVER_SUPPORT */
   LL_FOREACH_SAFE(session->delayqueue, q, tmp) {
     if (q->pdu->type==COAP_MESSAGE_CON && session->context && session->context->nack_handler)
       session->context->nack_handler(session, q->pdu, session->proto == COAP_PROTO_DTLS ? COAP_NACK_TLS_FAILED : COAP_NACK_NOT_DELIVERABLE, q->id);
@@ -216,10 +289,14 @@ void coap_session_mfree(coap_session_t *session) {
     LL_DELETE(session->lg_xmit, lq);
     coap_block_delete_lg_xmit(session, lq);
   }
+#if COAP_SERVER_SUPPORT
+  coap_lg_srcv_t *sq, *stmp;
+
   LL_FOREACH_SAFE(session->lg_srcv, sq, stmp) {
     LL_DELETE(session->lg_srcv, sq);
     coap_block_delete_lg_srcv(session, sq);
   }
+#endif /* COAP_SERVER_SUPPORT */
 }
 
 void coap_session_free(coap_session_t *session) {
@@ -229,20 +306,28 @@ void coap_session_free(coap_session_t *session) {
   if (session->ref)
     return;
   coap_session_mfree(session);
+#if COAP_SERVER_SUPPORT
   if (session->endpoint) {
     if (session->endpoint->sessions)
       SESSIONS_DELETE(session->endpoint->sessions, session);
-  } else if (session->context) {
+  } else
+#endif /* COAP_SERVER_SUPPORT */
+#if COAP_CLIENT_SUPPORT
+  if (session->context) {
     if (session->context->sessions)
       SESSIONS_DELETE(session->context->sessions, session);
   }
-  coap_log(LOG_DEBUG, "***%s: session closed\n", coap_session_str(session));
+#endif /* COAP_CLIENT_SUPPORT */
+  coap_delete_bin_const(session->last_token);
+  coap_log(LOG_DEBUG, "***%s: session %p: closed\n", coap_session_str(session),
+           (void *)session);
 
   coap_free_type(COAP_SESSION, session);
 }
 
-size_t coap_session_max_pdu_size(const coap_session_t *session) {
-  size_t max_with_header = (size_t)(session->mtu - session->tls_overhead);
+static size_t
+coap_session_max_pdu_size_internal(const coap_session_t *session,
+                                   size_t max_with_header) {
 #if COAP_DISABLE_TCP
   return max_with_header > 4 ? max_with_header - 4 : 0;
 #else /* !COAP_DISABLE_TCP */
@@ -262,11 +347,50 @@ size_t coap_session_max_pdu_size(const coap_session_t *session) {
 #endif /* !COAP_DISABLE_TCP */
 }
 
+size_t
+coap_session_max_pdu_rcv_size(const coap_session_t *session) {
+  if (session->csm_rcv_mtu)
+    return coap_session_max_pdu_size_internal(session,
+                                              (size_t)(session->csm_rcv_mtu));
+
+  return coap_session_max_pdu_size_internal(session,
+                              (size_t)(session->mtu - session->tls_overhead));
+}
+
+size_t
+coap_session_max_pdu_size(const coap_session_t *session) {
+  size_t max_with_header;
+
+#if COAP_CLIENT_SUPPORT
+  /*
+   * Delay if session->doing_first is set.
+   * E.g. Reliable and CSM not in yet for checking block support
+   */
+  coap_session_t *session_rw;
+
+  /*
+   * Need to do this to not get a compiler warning about const parameters
+   * but need to maintain source code backward compatibility
+   */
+  memcpy(&session_rw, &session, sizeof(session_rw));
+  if (coap_client_delay_first(session_rw) == 0) {
+    coap_log(LOG_DEBUG, "coap_client_delay_first: timeout\n");
+    /* Have to go with the defaults */
+  }
+#endif /* COAP_CLIENT_SUPPORT */
+
+  max_with_header = (size_t)(session->mtu - session->tls_overhead);
+
+  return coap_session_max_pdu_size_internal(session, max_with_header);
+}
+
 void coap_session_set_mtu(coap_session_t *session, unsigned mtu) {
 #if defined(WITH_CONTIKI) || defined(WITH_LWIP)
   if (mtu > COAP_MAX_MESSAGE_SIZE_TCP16 + 4)
     mtu = COAP_MAX_MESSAGE_SIZE_TCP16 + 4;
 #endif
+  if (mtu < 64)
+    mtu = 64;
   session->mtu = mtu;
   if (session->tls_overhead >= session->mtu) {
     session->tls_overhead = session->mtu;
@@ -278,10 +402,12 @@ ssize_t coap_session_send(coap_session_t *session, const uint8_t *data, size_t d
   ssize_t bytes_written;
 
   coap_socket_t *sock = &session->sock;
+#if COAP_SERVER_SUPPORT
   if (sock->flags == COAP_SOCKET_EMPTY) {
     assert(session->endpoint != NULL);
     sock = &session->endpoint->sock;
   }
+#endif /* COAP_SERVER_SUPPORT */
 
   bytes_written = coap_socket_send(sock, session, data, datalen);
   if (bytes_written == (ssize_t)datalen) {
@@ -300,7 +426,7 @@ ssize_t coap_session_write(coap_session_t *session, const uint8_t *data, size_t
   if (bytes_written > 0) {
     coap_ticks(&session->last_rx_tx);
     coap_log(LOG_DEBUG, "*  %s: sent %zd bytes\n",
-             coap_session_str(session), datalen);
+             coap_session_str(session), bytes_written);
   } else if (bytes_written < 0) {
     coap_log(LOG_DEBUG,  "*   %s: failed to send %zd bytes\n",
              coap_session_str(session), datalen );
@@ -320,13 +446,15 @@ coap_session_delay_pdu(coap_session_t *session, coap_pdu_t *pdu,
     node->session = NULL;
     node->t = 0;
   } else {
-    coap_queue_t *q = NULL;
-    /* Check that the same mid is not getting re-used in violation of RFC7252 */
-    LL_FOREACH(session->delayqueue, q) {
-      if (q->id == pdu->mid) {
-        coap_log(LOG_ERR, "**  %s: mid=0x%x: already in-use - dropped\n",
-                 coap_session_str(session), pdu->mid);
-        return COAP_INVALID_MID;
+    if (COAP_PROTO_NOT_RELIABLE(session->proto)) {
+      coap_queue_t *q = NULL;
+      /* Check same mid is not getting re-used in violation of RFC7252 */
+      LL_FOREACH(session->delayqueue, q) {
+        if (q->id == pdu->mid) {
+          coap_log(LOG_ERR, "**  %s: mid=0x%x: already in-use - dropped\n",
+                   coap_session_str(session), pdu->mid);
+          return COAP_INVALID_MID;
+        }
       }
     }
     node = coap_new_node();
@@ -359,10 +487,10 @@ void coap_session_send_csm(coap_session_t *session) {
     session->mtu = COAP_DEFAULT_MTU;  /* base value */
   pdu = coap_pdu_init(COAP_MESSAGE_CON, COAP_SIGNALING_CODE_CSM, 0, 20);
   if ( pdu == NULL
-    || coap_add_option(pdu, COAP_SIGNALING_OPTION_MAX_MESSAGE_SIZE,
+    || coap_add_option_internal(pdu, COAP_SIGNALING_OPTION_MAX_MESSAGE_SIZE,
          coap_encode_var_safe(buf, sizeof(buf),
-                                COAP_DEFAULT_MAX_PDU_RX_SIZE), buf) == 0
-    || coap_add_option(pdu, COAP_SIGNALING_OPTION_BLOCK_WISE_TRANSFER,
+                              session->context->csm_max_message_size), buf) == 0
+    || coap_add_option_internal(pdu, COAP_SIGNALING_OPTION_BLOCK_WISE_TRANSFER,
          coap_encode_var_safe(buf, sizeof(buf),
                                 0), buf) == 0
     || coap_pdu_encode_header(pdu, session->proto) == 0
@@ -370,8 +498,15 @@ void coap_session_send_csm(coap_session_t *session) {
     coap_session_disconnected(session, COAP_NACK_NOT_DELIVERABLE);
   } else {
     ssize_t bytes_written = coap_session_send_pdu(session, pdu);
-    if (bytes_written != (ssize_t)pdu->used_size + pdu->hdr_size)
+    if (bytes_written != (ssize_t)pdu->used_size + pdu->hdr_size) {
       coap_session_disconnected(session, COAP_NACK_NOT_DELIVERABLE);
+    } else {
+      session->csm_rcv_mtu = session->context->csm_max_message_size;
+      if (session->csm_rcv_mtu > COAP_BERT_BASE)
+        session->csm_bert_loc_support = 1;
+      else
+        session->csm_bert_loc_support = 0;
+    }
   }
   if (pdu)
     coap_delete_pdu(pdu);
@@ -402,6 +537,8 @@ void coap_session_connected(coap_session_t *session) {
              coap_session_str(session));
     if (session->state == COAP_SESSION_STATE_CSM)
       coap_handle_event(session->context, COAP_EVENT_SESSION_CONNECTED, session);
+    if (session->doing_first)
+      session->doing_first = 0;
   }
 
   session->state = COAP_SESSION_STATE_ESTABLISHED;
@@ -419,7 +556,7 @@ void coap_session_connected(coap_session_t *session) {
     ssize_t bytes_written;
     coap_queue_t *q = session->delayqueue;
     if (q->pdu->type == COAP_MESSAGE_CON && COAP_PROTO_NOT_RELIABLE(session->proto)) {
-      if (session->con_active >= COAP_DEFAULT_NSTART)
+      if (session->con_active >= COAP_NSTART(session))
         break;
       session->con_active++;
     }
@@ -460,7 +597,9 @@ void coap_session_disconnected(coap_session_t *session, coap_nack_reason_t reaso
 
   coap_log(LOG_DEBUG, "***%s: session disconnected (reason %d)\n",
            coap_session_str(session), reason);
+#if COAP_SERVER_SUPPORT
   coap_delete_observers( session->context, session );
+#endif /* COAP_SERVER_SUPPORT */
 
   if ( session->tls) {
     if (session->proto == COAP_PROTO_DTLS)
@@ -537,10 +676,13 @@ void coap_session_disconnected(coap_session_t *session, coap_nack_reason_t reaso
         state == COAP_SESSION_STATE_ESTABLISHED ?
         COAP_EVENT_SESSION_CLOSED : COAP_EVENT_SESSION_FAILED, session);
     }
+    if (session->doing_first)
+      session->doing_first = 0;
   }
 #endif /* !COAP_DISABLE_TCP */
 }
 
+#if COAP_SERVER_SUPPORT
 static void
 coap_make_addr_hash(coap_addr_hash_t *addr_hash, coap_proto_t proto,
                     const coap_addr_tuple_t *addr_info) {
@@ -604,11 +746,13 @@ coap_endpoint_get_session(coap_endpoint_t *endpoint,
 
   if (endpoint->context->max_idle_sessions > 0 &&
       num_idle >= endpoint->context->max_idle_sessions) {
+    coap_handle_event(oldest->context, COAP_EVENT_SERVER_SESSION_DEL, oldest);
     coap_session_free(oldest);
   }
   else if (oldest_hs) {
     coap_log(LOG_WARNING, "***%s: Incomplete session timed out\n",
              coap_session_str(oldest_hs));
+    coap_handle_event(oldest_hs->context, COAP_EVENT_SERVER_SESSION_DEL, oldest_hs);
     coap_session_free(oldest_hs);
   }
 
@@ -669,6 +813,7 @@ coap_endpoint_get_session(coap_endpoint_t *endpoint,
       return NULL;
     }
   }
+
   session = coap_make_session(endpoint->proto, COAP_SESSION_TYPE_SERVER,
                               &addr_hash, &packet->addr_info.local,
                               &packet->addr_info.remote,
@@ -681,8 +826,9 @@ coap_endpoint_get_session(coap_endpoint_t *endpoint,
       session->type = COAP_SESSION_TYPE_HELLO;
     }
     SESSIONS_ADD(endpoint->sessions, session);
-    coap_log(LOG_DEBUG, "***%s: new incoming session\n",
-             coap_session_str(session));
+    coap_log(LOG_DEBUG, "***%s: session %p: new incoming session\n",
+             coap_session_str(session), (void *)session);
+    coap_handle_event(session->context, COAP_EVENT_SERVER_SESSION_NEW, session);
   }
   return session;
 }
@@ -703,6 +849,7 @@ coap_session_new_dtls_session(coap_session_t *session,
   }
   return session;
 }
+#endif /* COAP_SERVER_SUPPORT */
 
 #ifdef COAP_EPOLL_SUPPORT
 static void
@@ -717,8 +864,12 @@ coap_epoll_ctl_add(coap_socket_t *sock,
   if (sock == NULL)
     return;
 
+#if COAP_SERVER_SUPPORT
   context = sock->session ? sock->session->context :
                             sock->endpoint ? sock->endpoint->context : NULL;
+#else /* ! COAP_SERVER_SUPPORT */
+  context = sock->session ? sock->session->context : NULL;
+#endif /* ! COAP_SERVER_SUPPORT */
   if (context == NULL)
     return;
 
@@ -737,6 +888,7 @@ coap_epoll_ctl_add(coap_socket_t *sock,
 }
 #endif /* COAP_EPOLL_SUPPORT */
 
+#if COAP_CLIENT_SUPPORT
 static coap_session_t *
 coap_session_create_client(
   coap_context_t *ctx,
@@ -782,11 +934,24 @@ coap_session_create_client(
   coap_session_reference(session);
 
   if (proto == COAP_PROTO_UDP || proto == COAP_PROTO_DTLS) {
+    coap_session_t *s, *rtmp;
     if (!coap_socket_connect_udp(&session->sock, local_if, server,
       proto == COAP_PROTO_DTLS ? COAPS_DEFAULT_PORT : COAP_DEFAULT_PORT,
       &session->addr_info.local, &session->addr_info.remote)) {
       goto error;
     }
+    /* Check that this is not a duplicate 4-tuple */
+    SESSIONS_ITER_SAFE(ctx->sessions, s, rtmp) {
+      if ((s->proto == COAP_PROTO_UDP || s->proto == COAP_PROTO_DTLS) &&
+          coap_address_equals(&session->addr_info.local,
+                              &s->addr_info.local) &&
+          coap_address_equals(&session->addr_info.remote,
+                              &s->addr_info.remote)) {
+        coap_log(LOG_WARNING, "***%s: session %p: duplicate - already exists\n",
+                 coap_session_str(session), (void *)session);
+        goto error;
+      }
+    }
 #if !COAP_DISABLE_TCP
   } else if (proto == COAP_PROTO_TCP || proto == COAP_PROTO_TLS) {
     if (!coap_socket_connect_tcp1(&session->sock, local_if, server,
@@ -809,10 +974,20 @@ coap_session_create_client(
   session->sock.flags |= COAP_SOCKET_NOT_EMPTY | COAP_SOCKET_WANT_READ;
   if (local_if)
     session->sock.flags |= COAP_SOCKET_BOUND;
+#if COAP_SERVER_SUPPORT
+  if (ctx->proxy_uri_resource)
+    session->proxy_session = 1;
+#endif /* COAP_SERVER_SUPPORT */
   SESSIONS_ADD(ctx->sessions, session);
   return session;
 
 error:
+  /*
+   * Need to add in the session as coap_session_release()
+   * will call SESSIONS_DELETE in coap_session_free().
+   */
+  if (session)
+    SESSIONS_ADD(ctx->sessions, session);
   coap_session_release(session);
   return NULL;
 }
@@ -839,6 +1014,12 @@ coap_session_connect(coap_session_t *session) {
     if (session->proto == COAP_PROTO_TCP || session->proto == COAP_PROTO_TLS) {
       if (session->sock.flags & COAP_SOCKET_WANT_CONNECT) {
         session->state = COAP_SESSION_STATE_CONNECTING;
+        if (session->state != COAP_SESSION_STATE_ESTABLISHED &&
+               session->state != COAP_SESSION_STATE_NONE &&
+               COAP_PROTO_RELIABLE(session->proto) &&
+               session->type == COAP_SESSION_TYPE_CLIENT) {
+          session->doing_first = 1;
+        }
       } else if (session->proto == COAP_PROTO_TLS) {
         int connected = 0;
         session->tls = coap_tls_new_client_session(session, &connected);
@@ -864,7 +1045,9 @@ coap_session_connect(coap_session_t *session) {
   coap_ticks(&session->last_rx_tx);
   return session;
 }
+#endif /* COAP_CLIENT_SUPPORT */
 
+#if COAP_SERVER_SUPPORT
 static coap_session_t *
 coap_session_accept(coap_session_t *session) {
 #if !COAP_DISABLE_TCP
@@ -894,17 +1077,20 @@ coap_session_accept(coap_session_t *session) {
 #endif /* COAP_DISABLE_TCP */
   return session;
 }
+#endif /* COAP_SERVER_SUPPORT */
 
+#if COAP_CLIENT_SUPPORT
 coap_session_t *coap_new_client_session(
   coap_context_t *ctx,
   const coap_address_t *local_if,
   const coap_address_t *server,
   coap_proto_t proto
 ) {
-  coap_session_t *session = coap_session_create_client(ctx, local_if, server, proto);
+  coap_session_t *session = coap_session_create_client(ctx, local_if, server,
+                                                       proto);
   if (session) {
-    coap_log(LOG_DEBUG, "***%s: new outgoing session\n",
-             coap_session_str(session));
+    coap_log(LOG_DEBUG, "***%s: session %p: created outgoing session\n",
+             coap_session_str(session), (void *)session);
     session = coap_session_connect(session);
   }
   return session;
@@ -922,6 +1108,7 @@ coap_session_t *coap_new_client_session_psk(
   coap_dtls_cpsk_t setup_data;
 
   memset (&setup_data, 0, sizeof(setup_data));
+  setup_data.version = COAP_DTLS_CPSK_SETUP_VERSION;
 
   if (identity) {
     setup_data.psk_info.identity.s = (const uint8_t *)identity;
@@ -961,7 +1148,7 @@ coap_session_t *coap_new_client_session_psk2(
       return NULL;
     }
   }
-  else if (coap_dtls_is_supported()) {
+  else if (coap_dtls_is_supported() || coap_tls_is_supported()) {
     coap_log(LOG_WARNING, "Identity (PSK) not defined\n");
     coap_session_release(session);
     return NULL;
@@ -976,13 +1163,13 @@ coap_session_t *coap_new_client_session_psk2(
       return NULL;
     }
   }
-  else if (coap_dtls_is_supported()) {
+  else if (coap_dtls_is_supported() || coap_tls_is_supported()) {
     coap_log(LOG_WARNING, "Pre-shared key (PSK) not defined\n");
     coap_session_release(session);
     return NULL;
   }
 
-  if (coap_dtls_is_supported()) {
+  if (coap_dtls_is_supported() || coap_tls_is_supported()) {
     if (!coap_dtls_context_set_cpsk(ctx, setup_data)) {
       coap_session_release(session);
       return NULL;
@@ -992,8 +1179,10 @@ coap_session_t *coap_new_client_session_psk2(
            coap_session_str(session));
   return coap_session_connect(session);
 }
+#endif /* ! COAP_CLIENT_SUPPORT */
 
-int coap_session_refresh_psk_hint(coap_session_t *session,
+int
+coap_session_refresh_psk_hint(coap_session_t *session,
   const coap_bin_const_t *psk_hint
 ) {
   /* We may be refreshing the hint with the same hint */
@@ -1022,7 +1211,8 @@ int coap_session_refresh_psk_hint(coap_session_t *session,
   return 1;
 }
 
-int coap_session_refresh_psk_key(coap_session_t *session,
+int
+coap_session_refresh_psk_key(coap_session_t *session,
   const coap_bin_const_t *psk_key
 ) {
   /* We may be refreshing the key with the same key */
@@ -1050,12 +1240,56 @@ int coap_session_refresh_psk_key(coap_session_t *session,
   return 1;
 }
 
+int
+coap_session_refresh_psk_identity(coap_session_t *session,
+  const coap_bin_const_t *psk_identity
+) {
+  /* We may be refreshing the identity with the same identity */
+  coap_bin_const_t *old_psk_identity = session->psk_identity;
+
+  if (psk_identity && psk_identity->s) {
+    if (session->psk_identity) {
+      if (coap_binary_equal(session->psk_identity, psk_identity))
+        return 1;
+    }
+    session->psk_identity = coap_new_bin_const(psk_identity->s,
+                                               psk_identity->length);
+    if (!session->psk_identity) {
+      coap_log(LOG_ERR, "No memory to store pre-shared key identity (PSK)\n");
+      if (old_psk_identity)
+        coap_delete_bin_const(old_psk_identity);
+      return 0;
+    }
+  }
+  else {
+    session->psk_identity = NULL;
+  }
+  if (old_psk_identity)
+    coap_delete_bin_const(old_psk_identity);
+
+  return 1;
+}
+
+#if COAP_SERVER_SUPPORT
 const coap_bin_const_t *
 coap_session_get_psk_hint(const coap_session_t *session) {
   if (session)
     return session->psk_hint;
   return NULL;
 }
+#endif /* COAP_SERVER_SUPPORT */
+
+const coap_bin_const_t *
+coap_session_get_psk_identity(const coap_session_t *session) {
+  const coap_bin_const_t *psk_identity = NULL;
+  if (session) {
+    psk_identity = session->psk_identity;
+    if (psk_identity == NULL) {
+      psk_identity = &session->cpsk_setup_data.psk_info.identity;
+    }
+  }
+  return psk_identity;
+}
 
 const coap_bin_const_t *
 coap_session_get_psk_key(const coap_session_t *session) {
@@ -1064,6 +1298,7 @@ coap_session_get_psk_key(const coap_session_t *session) {
   return NULL;
 }
 
+#if COAP_CLIENT_SUPPORT
 coap_session_t *coap_new_client_session_pki(
   coap_context_t *ctx,
   const coap_address_t *local_if,
@@ -1073,7 +1308,7 @@ coap_session_t *coap_new_client_session_pki(
 ) {
   coap_session_t *session;
 
-  if (coap_dtls_is_supported()) {
+  if (coap_dtls_is_supported() || coap_tls_is_supported()) {
     if (!setup_data) {
       return NULL;
     } else {
@@ -1091,7 +1326,7 @@ coap_session_t *coap_new_client_session_pki(
     return NULL;
   }
 
-  if (coap_dtls_is_supported()) {
+  if (coap_dtls_is_supported() || coap_tls_is_supported()) {
     /* we know that setup_data is not NULL */
     if (!coap_dtls_context_set_pki(ctx, setup_data, COAP_DTLS_ROLE_CLIENT)) {
       coap_session_release(session);
@@ -1102,7 +1337,9 @@ coap_session_t *coap_new_client_session_pki(
            coap_session_str(session));
   return coap_session_connect(session);
 }
+#endif /* ! COAP_CLIENT_SUPPORT */
 
+#if COAP_SERVER_SUPPORT
 coap_session_t *coap_new_server_session(
   coap_context_t *ctx,
   coap_endpoint_t *ep
@@ -1131,16 +1368,28 @@ coap_session_t *coap_new_server_session(
 #endif /* COAP_EPOLL_SUPPORT */
   SESSIONS_ADD(ep->sessions, session);
   if (session) {
-    coap_log(LOG_DEBUG, "***%s: new incoming session\n",
-             coap_session_str(session));
+    coap_log(LOG_DEBUG, "***%s: session %p: new incoming session\n",
+             coap_session_str(session), (void *)session);
+    /* Returned session may already have been released and is now NULL */
     session = coap_session_accept(session);
+    if(session) {
+      coap_handle_event(session->context, COAP_EVENT_SERVER_SESSION_NEW, session);
+    }
   }
   return session;
 
 error:
-  coap_session_free(session);
+  /*
+   * Need to add in the session as coap_session_release()
+   * will call SESSIONS_DELETE in coap_session_free().
+   */
+  if (session) {
+    SESSIONS_ADD(ep->sessions, session);
+    coap_session_free(session);
+  }
   return NULL;
 }
+#endif /* COAP_SERVER_SUPPORT */
 
 void
 coap_session_init_token(coap_session_t *session, size_t len,
@@ -1194,15 +1443,21 @@ coap_session_get_type(const coap_session_t *session) {
   return 0;
 }
 
+#if COAP_CLIENT_SUPPORT
 int
 coap_session_set_type_client(coap_session_t *session) {
+#if COAP_SERVER_SUPPORT
   if (session && session->type == COAP_SESSION_TYPE_SERVER) {
     coap_session_reference(session);
     session->type = COAP_SESSION_TYPE_CLIENT;
     return 1;
   }
+#else /* ! COAP_SERVER_SUPPORT */
+  (void)session;
+#endif /* ! COAP_SERVER_SUPPORT */
   return 0;
 }
+#endif /* COAP_CLIENT_SUPPORT */
 
 coap_session_state_t
 coap_session_get_state(const coap_session_t *session) {
@@ -1225,6 +1480,7 @@ void *coap_session_get_tls(const coap_session_t *session,
 }
 
 #ifndef WITH_LWIP
+#if COAP_SERVER_SUPPORT
 coap_endpoint_t *
 coap_new_endpoint(coap_context_t *context, const coap_address_t *listen_addr, coap_proto_t proto) {
   coap_endpoint_t *ep = NULL;
@@ -1354,6 +1610,7 @@ coap_free_endpoint(coap_endpoint_t *ep) {
     coap_mfree_endpoint(ep);
   }
 }
+#endif /* COAP_SERVER_SUPPORT */
 #endif /* WITH_LWIP */
 
 coap_session_t *
@@ -1361,12 +1618,16 @@ coap_session_get_by_peer(const coap_context_t *ctx,
   const coap_address_t *remote_addr,
   int ifindex) {
   coap_session_t *s, *rtmp;
-  coap_endpoint_t *ep;
+#if COAP_CLIENT_SUPPORT
   SESSIONS_ITER(ctx->sessions, s, rtmp) {
     if (s->ifindex == ifindex && coap_address_equals(&s->addr_info.remote,
                                                      remote_addr))
       return s;
   }
+#endif /* COAP_CLIENT_SUPPORT */
+#if COAP_SERVER_SUPPORT
+  coap_endpoint_t *ep;
+
   LL_FOREACH(ctx->endpoint, ep) {
     SESSIONS_ITER(ep->sessions, s, rtmp) {
       if (s->ifindex == ifindex && coap_address_equals(&s->addr_info.remote,
@@ -1374,6 +1635,7 @@ coap_session_get_by_peer(const coap_context_t *ctx,
         return s;
     }
   }
+#endif /* COAP_SERVER_SUPPORT */
   return NULL;
 }
 
@@ -1416,6 +1678,7 @@ const char *coap_session_str(const coap_session_t *session) {
   return szSession;
 }
 
+#if COAP_SERVER_SUPPORT
 const char *coap_endpoint_str(const coap_endpoint_t *endpoint) {
   static char szEndpoint[128];
   char *p = szEndpoint, *end = szEndpoint + sizeof(szEndpoint);
@@ -1436,5 +1699,11 @@ const char *coap_endpoint_str(const coap_endpoint_t *endpoint) {
 
   return szEndpoint;
 }
-
+#endif /* COAP_SERVER_SUPPORT */
+#ifdef COAP_CLIENT_SUPPORT
+void
+coap_session_set_no_observe_cancel(coap_session_t *session) {
+  session->no_observe_cancel = 1;
+}
+#endif /* COAP_CLIENT_SUPPORT */
 #endif  /* COAP_SESSION_C_ */
