@@ -2,7 +2,7 @@
  * coap_tinydtls.c -- Datagram Transport Layer Support for libcoap with tinydtls
  *
  * Copyright (C) 2016-2020 Olaf Bergmann <bergmann@tzi.org>
- * Copyright (C) 2020 Jon Shallow <supjps-libcoap@jpshallow.com>
+ * Copyright (C) 2020-2022 Jon Shallow <supjps-libcoap@jpshallow.com>
  *
  * SPDX-License-Identifier: BSD-2-Clause
  *
@@ -10,6 +10,11 @@
  * of use.
  */
 
+/**
+ * @file coap_tinydtls.c
+ * @brief TinyDTLS specific handling functions
+ */
+
 #include "coap3/coap_internal.h"
 
 #ifdef HAVE_LIBTINYDTLS
@@ -22,9 +27,9 @@
 #undef PACKAGE_URL
 #undef PACKAGE_VERSION
 
-#include <tinydtls.h>
-#include <dtls.h>
-#include <dtls_debug.h>
+#include <tinydtls/tinydtls.h>
+#include <tinydtls/dtls.h>
+#include <tinydtls/dtls_debug.h>
 
 typedef struct coap_tiny_context_t {
   struct dtls_context_t *dtls_context;
@@ -206,13 +211,17 @@ get_psk_info(struct dtls_context_t *dtls_context,
   coap_context_t *coap_context = t_context ? t_context->coap_context : NULL;
   coap_session_t *coap_session;
   int fatal_error = DTLS_ALERT_INTERNAL_ERROR;
-  size_t identity_length;
-  uint8_t psk[128];
-  size_t psk_len = 0;
   coap_address_t remote_addr;
+#if COAP_CLIENT_SUPPORT
   coap_dtls_cpsk_t *setup_cdata;
+  const coap_bin_const_t *psk_identity;
+  const coap_dtls_cpsk_info_t *cpsk_info;
+#endif /* COAP_CLIENT_SUPPORT */
+  const coap_bin_const_t *psk_key;
+#if COAP_SERVER_SUPPORT
   coap_dtls_spsk_t *setup_sdata;
-  coap_bin_const_t temp;
+  const coap_bin_const_t *psk_hint;
+#endif /* COAP_SERVER_SUPPORT */
 
   assert(coap_context);
   get_session_addr(dtls_session, &remote_addr);
@@ -225,111 +234,147 @@ get_psk_info(struct dtls_context_t *dtls_context,
   switch (type) {
   case DTLS_PSK_IDENTITY:
 
-    if (!coap_context || !coap_context->get_client_psk ||
-        coap_session->type != COAP_SESSION_TYPE_CLIENT)
+#if COAP_CLIENT_SUPPORT
+    if (coap_session->type != COAP_SESSION_TYPE_CLIENT)
       goto error;
 
     setup_cdata = &coap_session->cpsk_setup_data;
 
+    coap_bin_const_t temp;
     temp.s = id;
     temp.length = id_len;
     coap_session_refresh_psk_hint(coap_session, &temp);
 
-    coap_log(LOG_DEBUG, "got psk_identity_hint: '%.*s'\n", (int)id_len, id ? (const char*)id : "");
+    coap_log(LOG_DEBUG, "got psk_identity_hint: '%.*s'\n", (int)id_len,
+             id ? (const char*)id : "");
 
     if (setup_cdata->validate_ih_call_back) {
       coap_str_const_t lhint;
+
       lhint.length = id_len;
       lhint.s = id;
-      const coap_dtls_cpsk_info_t *psk_info =
-               setup_cdata->validate_ih_call_back(&lhint,
-                                                  coap_session,
-                                                  setup_cdata->ih_call_back_arg);
-
-      if (psk_info == NULL)
-        return 0;
-      if (psk_info->identity.length >= result_length)
-        return 0;
-      if (psk_info->key.length > sizeof(psk))
-        return 0;
-
-      if (coap_session->psk_identity) {
-        coap_delete_bin_const(coap_session->psk_identity);
+      cpsk_info =
+             setup_cdata->validate_ih_call_back(&lhint,
+                                                coap_session,
+                                                setup_cdata->ih_call_back_arg);
+      if (cpsk_info) {
+        psk_identity = &cpsk_info->identity;
+        coap_session_refresh_psk_identity(coap_session, &cpsk_info->identity);
+        coap_session_refresh_psk_key(coap_session, &cpsk_info->key);
+      }
+      else {
+        psk_identity = NULL;
       }
-      identity_length = psk_info->identity.length;
-      coap_session->psk_identity = coap_new_bin_const(psk_info->identity.s, identity_length);
-      memcpy(result, psk_info->identity.s, identity_length);
-      result[identity_length] = '\000';
-
-      coap_session_refresh_psk_key(coap_session, &psk_info->key);
-
-      return identity_length;
     }
-
-    identity_length = 0;
-    /* result_length is max size of the returned identity */
-    psk_len = coap_context->get_client_psk(coap_session, (const uint8_t*)id, id_len, (uint8_t*)result, &identity_length, result_length, psk, sizeof(psk));
-    if (!psk_len) {
-      coap_log(LOG_WARNING, "no PSK identity for given realm or buffer too small\n");
+    else {
+      psk_identity = coap_get_session_client_psk_identity(coap_session);
+    }
+    if (psk_identity == NULL) {
+      coap_log(LOG_WARNING, "no PSK identity given\n");
       fatal_error = DTLS_ALERT_CLOSE_NOTIFY;
       goto error;
     }
-    return (int)identity_length;
+    if (psk_identity->length > result_length) {
+      coap_log(LOG_WARNING,
+               "psk_identity too large, truncated to %zd bytes\n",
+               result_length);
+    }
+    else {
+      /* Reduce to match */
+      result_length = psk_identity->length;
+    }
+    memcpy(result, psk_identity->s, result_length);
+    return result_length;
+#else /* ! COAP_CLIENT_SUPPORT */
+    return 0;
+#endif /* ! COAP_CLIENT_SUPPORT */
 
   case DTLS_PSK_KEY:
+#if COAP_CLIENT_SUPPORT
     if (coap_session->type == COAP_SESSION_TYPE_CLIENT) {
-      if (!coap_context || !coap_context->get_client_psk)
-        goto error;
-      identity_length = 0;
-      /* Use psk[] as a scratch area for returning the unused identity */
-      psk_len = coap_context->get_client_psk(coap_session, (const uint8_t*)id, id_len, psk, &identity_length, sizeof(psk), result, result_length);
-      if (!psk_len) {
-        coap_log(LOG_WARNING, "no pre-shared key for given realm or buffer too small\n");
+      psk_key = coap_get_session_client_psk_key(coap_session);
+      if (psk_key == NULL) {
+        coap_log(LOG_WARNING, "no PSK key given\n");
         fatal_error = DTLS_ALERT_CLOSE_NOTIFY;
         goto error;
       }
-      return (int)psk_len;
+      if (psk_key->length > result_length) {
+        coap_log(LOG_WARNING,
+                 "psk_key too large, truncated to %zd bytes\n",
+                 result_length);
+      }
+      else {
+        /* Reduce to match */
+        result_length = psk_key->length;
+      }
+      memcpy(result, psk_key->s, result_length);
+      return result_length;
     }
-    if (coap_context->get_server_psk) {
+#endif /* COAP_CLIENT_SUPPORT */
+#if COAP_SERVER_SUPPORT
+    if (coap_session->type != COAP_SESSION_TYPE_CLIENT) {
+      coap_bin_const_t lidentity;
+
+      lidentity.length = id ? id_len : 0;
+      lidentity.s = id ? (const uint8_t*)id : (const uint8_t *)"";
       setup_sdata = &coap_session->context->spsk_setup_data;
 
-      if (!id)
-        id = (const uint8_t *)"";
-
       /* Track the Identity being used */
-      if (coap_session->psk_identity)
-        coap_delete_bin_const(coap_session->psk_identity);
-      coap_session->psk_identity = coap_new_bin_const(id, id_len);
+      coap_session_refresh_psk_identity(coap_session, &lidentity);
 
       coap_log(LOG_DEBUG, "got psk_identity: '%.*s'\n",
-               (int)id_len, id);
+               (int)lidentity.length, lidentity.s);
 
       if (setup_sdata->validate_id_call_back) {
-        coap_bin_const_t lidentity;
-        lidentity.length = id_len;
-        lidentity.s = (const uint8_t*)id;
-        const coap_bin_const_t *psk_key =
-                 setup_sdata->validate_id_call_back(&lidentity,
-                                                    coap_session,
-                                                    setup_sdata->id_call_back_arg);
-
-        if (psk_key == NULL)
-          return 0;
-        if (psk_key->length > result_length)
-          return 0;
-        memcpy(result, psk_key->s, psk_key->length);
-        coap_session_refresh_psk_key(coap_session, psk_key);
-        return psk_key->length;
+        psk_key =
+              setup_sdata->validate_id_call_back(&lidentity,
+                                                 coap_session,
+                                                 setup_sdata->id_call_back_arg);
+      }
+      else {
+        psk_key = coap_get_session_server_psk_key(coap_session);
       }
 
-      return (int)coap_context->get_server_psk(coap_session, (const uint8_t*)id, id_len, (uint8_t*)result, result_length);
+      if (psk_key == NULL) {
+        coap_log(LOG_WARNING, "no PSK key given\n");
+        return 0;
+      }
+      if (setup_sdata->validate_id_call_back)
+        coap_session_refresh_psk_key(coap_session, psk_key);
+      if (psk_key->length > result_length) {
+        coap_log(LOG_WARNING,
+                 "psk_key too large, truncated to %zd bytes\n",
+                 result_length);
+      }
+      else {
+        /* Reduce to match */
+        result_length = psk_key->length;
+      }
+      memcpy(result, psk_key->s, result_length);
+      return result_length;
     }
+#endif /* COAP_SERVER_SUPPORT */
     return 0;
 
   case DTLS_PSK_HINT:
-    if (coap_context->get_server_hint)
-      return (int)coap_context->get_server_hint(coap_session, (uint8_t *)result, result_length);
+#if COAP_SERVER_SUPPORT
+    psk_hint = coap_get_session_server_psk_hint(coap_session);
+    if (psk_hint == NULL)
+      return 0;
+    if (psk_hint->length > result_length) {
+      coap_log(LOG_WARNING,
+               "psk_hint too large, truncated to %zd bytes\n",
+               result_length);
+    }
+    else {
+      /* Reduce to match */
+      result_length = psk_hint->length;
+    }
+    memcpy(result, psk_hint->s, result_length);
+    return result_length;
+#else /* COAP_SERVER_SUPPORT */
     return 0;
+#endif /* COAP_SERVER_SUPPORT */
 
   default:
     coap_log(LOG_WARNING, "unsupported request type: %d\n", type);
@@ -489,10 +534,13 @@ coap_dtls_new_session(coap_session_t *session) {
   return dtls_session;
 }
 
+#if COAP_SERVER_SUPPORT
 void *coap_dtls_new_server_session(coap_session_t *session) {
   return coap_dtls_new_session(session);
 }
+#endif /* COAP_SERVER_SUPPORT */
 
+#if COAP_CLIENT_SUPPORT
 void *coap_dtls_new_client_session(coap_session_t *session) {
   dtls_peer_t *peer;
   coap_tiny_context_t *t_context = (coap_tiny_context_t *)session->context->dtls_context;
@@ -522,6 +570,7 @@ void *coap_dtls_new_client_session(coap_session_t *session) {
 
   return dtls_session;
 }
+#endif /* COAP_CLIENT_SUPPORT */
 
 void
 coap_dtls_session_update_mtu(coap_session_t *session) {
@@ -605,9 +654,14 @@ coap_tick_t coap_dtls_get_timeout(coap_session_t *session, coap_tick_t now) {
   return 0;
 }
 
-void coap_dtls_handle_timeout(coap_session_t *session) {
+/*
+ * return 1 timed out
+ *        0 still timing out
+ */
+int
+coap_dtls_handle_timeout(coap_session_t *session) {
   (void)session;
-  return;
+  return 0;
 }
 
 int
@@ -644,6 +698,7 @@ coap_dtls_receive(coap_session_t *session,
   return err;
 }
 
+#if COAP_SERVER_SUPPORT
 int
 coap_dtls_hello(coap_session_t *session,
   const uint8_t *data,
@@ -670,6 +725,7 @@ coap_dtls_hello(coap_session_t *session,
   }
   return res;
 }
+#endif /* COAP_SERVER_SUPPORT */
 
 unsigned int coap_dtls_get_overhead(coap_session_t *session) {
   (void)session;
@@ -1139,6 +1195,7 @@ coap_dtls_context_set_pki_root_cas(coap_context_t *ctx COAP_UNUSED,
   return 0;
 }
 
+#if COAP_CLIENT_SUPPORT
 int
 coap_dtls_context_set_cpsk(coap_context_t *coap_context COAP_UNUSED,
   coap_dtls_cpsk_t *setup_data
@@ -1148,7 +1205,9 @@ coap_dtls_context_set_cpsk(coap_context_t *coap_context COAP_UNUSED,
 
   return 1;
 }
+#endif /* COAP_CLIENT_SUPPORT */
 
+#if COAP_SERVER_SUPPORT
 int
 coap_dtls_context_set_spsk(coap_context_t *coap_context COAP_UNUSED,
   coap_dtls_spsk_t *setup_data
@@ -1163,6 +1222,7 @@ coap_dtls_context_set_spsk(coap_context_t *coap_context COAP_UNUSED,
 
   return 1;
 }
+#endif /* COAP_SERVER_SUPPORT */
 
 int
 coap_dtls_context_check_keys_enabled(coap_context_t *ctx COAP_UNUSED)
@@ -1171,13 +1231,17 @@ coap_dtls_context_check_keys_enabled(coap_context_t *ctx COAP_UNUSED)
 }
 
 #if !COAP_DISABLE_TCP
+#if COAP_CLIENT_SUPPORT
 void *coap_tls_new_client_session(coap_session_t *session COAP_UNUSED, int *connected COAP_UNUSED) {
   return NULL;
 }
+#endif /* COAP_CLIENT_SUPPORT */
 
+#if COAP_SERVER_SUPPORT
 void *coap_tls_new_server_session(coap_session_t *session COAP_UNUSED, int *connected COAP_UNUSED) {
   return NULL;
 }
+#endif /* COAP_SERVER_SUPPORT */
 
 void coap_tls_free_session(coap_session_t *coap_session COAP_UNUSED) {
 }
@@ -1197,6 +1261,7 @@ ssize_t coap_tls_read(coap_session_t *session COAP_UNUSED,
 }
 #endif /* !COAP_DISABLE_TCP */
 
+#if COAP_SERVER_SUPPORT
 coap_digest_ctx_t *
 coap_digest_setup(void) {
   dtls_sha256_ctx *digest_ctx = coap_malloc(sizeof(dtls_sha256_ctx));
@@ -1230,6 +1295,7 @@ coap_digest_final(coap_digest_ctx_t *digest_ctx,
   coap_digest_free(digest_ctx);
   return 1;
 }
+#endif /* COAP_SERVER_SUPPORT */
 
 #else /* !HAVE_LIBTINYDTLS */
 
